/* 
========================================
   ** SupCom MaxTool - Jan Van Der Weg **
          www.janvanderweg.com

	Many thanks to:
		Chris Cookson for his scripts, they taught me a lot!
		der_ton from doom3world, without his scripts correct bone creation would have taken a lot longer!
		GH33DA for explaining what the transform matrixes (matrices?) are!
	
	SupCom Maxtool:
	v0.49a
	-Fixed bone creation. It should all be correct now. 
	-Optimized a lot of the script and fixed some potential issues.
	-Still some small things to add but the base import functions are finished.
	v0.4a
	-Planning on merging scm/sca import and export into a single maxscript: SupCom MaxTool
	-UI layout change:
		-Added option to load reference points as dummies or bones.
		-Added progress bar
	-Bones are now max bone objects.
	-Bone location is still off with the spiderbot. I think this is still a bug :(
	-Apply sca button has no functionality yet, sca import script has not been merged in yet.
========================================
*/
utility SCMTool "SupCom MaxTool" silentErrors:false
(
	--Rollout
	group "About:"
	(
		bitmap the_bmp fileName:"supcom.bmp" 
		label Label1 "SupCom MaxTool v0.49a"
		label Label2 "www.janvanderweg.com"
	)
	
	group "Import SCM:"
	(
		--label Label3 "Load bones..."
		--radiobuttons radiobones labels:#("Only weighted", "All bones")
		--checkbox fixrotcheck "Fix model rotation" tristate:1
		checkbox loadrefpoints "Load ref. bones as:"
		radiobuttons refpointsas labels:#("Dummies","Bones")
		checkbox fixrotcheck "Fix model rotation" tristate:1

		button importButton "Import .SCM"	tooltip:"Import a Supreme Command An RTS From Chris Taylor model... ^_^"
	)
	group "Import SCA:"
	(
		button applySCA "Apply .SCA" tooltip:"Load and apply animations from an .sca file... ^_^"
	)
	
	---------------
	-- Utilities --
	---------------
	-- Replacement for doesFileExist which seems to give false positives
	fn DoesFileReallyExist filename=
	(
		local temp = fopen filename "rb"
		if temp != undefined then
		(
			fclose temp
			true
		)
		else
		(
			false
		)
	)
	-- Read a string of fixedLen chars
	fn ReadFixedString bstream fixedLen junk=
	(
		local str = ""
		for i = 1 to fixedLen do
		(
			local ch = ReadByte bstream #unsigned
			--format "Read string: %\n" ch
			if junk == true then
			(
				if ch == 0 then
				(
					str += bit.intAsChar ch
					-- Ignore junk
					SkipBytes bstream (fixedLen - i)
					exit
				)
				else 
				(
					str += bit.intAsChar ch
				)
			)
			--str += bit.intAsChar ch
			else if junk == false then
			(
				if ch == 0 then 
				(
					--replace with a tab "\t"
					str += "\t"

				)
				else
				(
					str += bit.intAsChar ch
				)
			)
		)
		str
	)
	--triangle struct
	struct SCM_Tris
	(
		triindex, --3 unsigned shorts 
		
		fn LoadFromStream bstream=
		(
			local x = Readshort bstream #unsigned
			local y = Readshort bstream #unsigned
			local z = Readshort bstream #unsigned
			triindex = [x, y, z]
		)
	)
	--vertex struct
	struct SCM_Vert
	(
		Position, --float vector(x,y,z) position relative to the local origion of the mesh
		--Tangent Space
		Normal, -- float vector(x,y,z)
		Tangent, -- float vector(x,y,z)
		Binormal, -- float vector(x,y,z)
		--supports two sets of uv coordinates (awesome!)
		uv1, --floatvector(u,v)
		uv2, --floatvector(u,v)
		boneindex, -- array of 4 chars - verts can be weighted to up to 4 bones. Supcom only expects boneindex[0] (non organic models)
	
		fn LoadFromStream bstream=
		(
				local x = ReadFloat bstream
				local y = ReadFloat bstream
				local z = ReadFloat bstream
				Position = [x, y, z]
				local x = ReadFloat bstream
				local y = ReadFloat bstream
				local z = ReadFloat bstream
				Normal = [x, y ,z]
				local x = ReadFloat bstream
				local y = ReadFloat bstream
				local z = ReadFloat bstream
				Tangent = [x, y, z]
				local x = ReadFloat bstream
				local y = ReadFloat bstream
				local z = ReadFloat bstream
				Binormal = [x, y, z]
				--hardcode the w component
				local u = ReadFloat bstream
				local v = ( 1 - ReadFloat bstream )
				uv1 = [u, v, 0]
				--hardcode the w component
				local u = ReadFloat bstream
				local v = ( 1 - ReadFloat bstream )
				uv2 = [u, v, 0]
				local x = Readbyte bstream #unsigned
				local y = Readbyte bstream #unsigned
				local z = Readbyte bstream #unsigned
				local w = Readbyte bstream #unsigned
				boneindex = [x, y, z, w]
		)	
	)
	--bone struct
	struct SCM_Bone
	(
	    RestPoseInverse, 	--4x4 Matrix with row major 
		Position,				--vector(x,z,y) pos relative to parent bone
		Rotation,				--quaternion(w,x,y,z) rot relative to parent bone
		nameoffset,				--long
		parentboneindex,		--long
		reserved0,				--long				
		reserved1,				--long
		Rotation2,
		rotquat,
			
		fn LoadFromStream bstream=
		(
			
			row1_point3 = point3 ( ReadFloat bstream ) ( ReadFloat bstream ) ( ReadFloat bstream )
			w = ReadFloat bstream
			row2_point3 = point3 ( ReadFloat bstream ) ( ReadFloat bstream ) ( ReadFloat bstream )
			w = ReadFloat bstream
			row3_point3 = point3 ( ReadFloat bstream ) ( ReadFloat bstream ) ( ReadFloat bstream )
			w = ReadFloat bstream
			row4_point3 = point3 ( ReadFloat bstream ) ( ReadFloat bstream ) ( ReadFloat bstream )
			w = ReadFloat bstream
			
			RestPoseInverse = matrix3 row1_point3 row2_point3 row3_point3 row4_point3
					
			x = ReadFloat bstream
			y = ReadFloat bstream
			z = ReadFloat bstream
			Position = [x, y, z]
			w = ReadFloat bstream
			x = ReadFloat bstream
			y = ReadFloat bstream
			z = ReadFloat bstream
			rotquat = Quat x y z w
			nameoffset = ReadLong bstream #unsigned
			parentboneindex = ( ReadLong bstream #unsigned + 1 )
			reserved0 = ReadLong bstream #unsigned
			reserved1 = ReadLong bstream #unsigned
		),		
	)
	struct SCM_Header
	(
		fourcc,			-- "MODL"
		version,		-- "5"
		boneoffset,		
		wbone,			--numer of bones influencing verts
		vertoffset,
		evertoffset,	--not used in supcom1.0, omitted
		vertcount,		--vertex count
		indexoffset,		
		indexcount,		--tricount = indexcount/3
		infooffset,
		infocount,
		totalbones,		--total amount of bones (weighted bones and tags)
		tricount,
	
		fn LoadFromStream bstream=
		(
			fourcc 		= ReadFixedString bstream 4 true
			version 	= ReadLong bstream #unsigned
			boneoffset 	= ReadLong bstream #unsigned
			wbone 		= ReadLong bstream #unsigned
			vertoffset 	= ReadLong bstream #unsigned
			evertoffset = ReadLong bstream #unsigned
			vertcount 	= ReadLong bstream #unsigned
			indexoffset = ReadLong bstream #unsigned
			indexcount 	= ReadLong bstream #unsigned
			infooffset 	= ReadLong bstream #unsigned
			infocount 	= ReadLong bstream #unsigned
			totalbones 	= ReadLong bstream #unsigned
			tricount 	= ( indexcount / 3 )
		),

		fn IsValid=
		(
			if fourcc != "MODL" then return false
			if version != 5 then
			(
				ShowError "Incorrect version! Found " + version + " but should be 5"
				return false
			)
			true
		),
		
		fn DebugOut=
		(
			format "SCM_Header\n----------\n"
			format "Marker: %, version: %\n" fourcc version
			format "Weighted bones: %, Total bones: %\n" wbone totalbones
			format "Vertex count: %, Trianlge count: %\n" vertcount ( indexcount / 3 )
			format "bOffset: %, vOffset: %, tOffset: %, iOffset %\n" boneoffset vertoffset indexoffset infooffset
			ok
		)
	)	
	
	-- The main SCM loader object
	struct SCM_Loader
	(
		bstream,	--bitsream
		header,		--file header
		bonelist=#(), --array of bone structs
		bonenames=#(), --array of bone names
		vertlist=#(), --array of vert structs
		trislist=#(), --array of tris structs
		vertweight=#(), --array that holds which bone is linked to which vert
		boneweight=#(), --array that holds the bone number
		
		fn Open filename=
		(
			g_modelFilename = filename
			g_modelName = getFilenameFile filename
			g_modelPath = getFileNamePath filename
			bstream = fopen filename "rb"
			ok
		),
		
		fn Close=
		(
			fclose bstream
			--print "Closed filestream"
			ok
		),
		-- Create the mesh of the model and its associated data (uv, bone weights)
		fn CreateModel=
		(
			progressUpdate (100.0*7/16)
			format "Creating model\n"
			vert_array=#() -- array that will hold the vertex information
			face_array=#() -- array that will hold the triangle list information
			tvert_array=#() -- array that will hold vertex texture coordinates
			-- build vert array
			for i = 1 to header.vertcount do
			(
				local currentvert = vertlist[i].position
				append vert_array currentvert 
				--use first set of uv's
			    local currenttvert = vertlist[i].uv1
				append tvert_array currenttvert
			)
			progressUpdate (100.0*8/16)
			-- build triangle list
			for i = 1 to header.tricount do
			(
				local currentface = trislist[i].triindex
				currentface[1] = currentface[1] + 1
				currentface[2] = currentface[2] + 1
				currentface[3] = currentface[3] + 1
				append face_array currentface
			)
			progressUpdate (100.0*9/16)
			--create a new mesh object
			new_mesh = mesh vertices:vert_array faces:face_array tverts:tvert_array
			--fix tverts
			fix_mesh = new_mesh.mesh
			BuildTVFaces fix_mesh
			for i = 1 to fix_mesh.numfaces do (setTVFace fix_mesh i (getFace fix_mesh i))
			
			--rebuild the mesh object
			update fix_mesh
			progressUpdate (100.0*10/16)
			--add a skinmodifier to the mesh
			addmodifier new_mesh (skin())
			select new_mesh
			max modify mode
			skinmod=new_mesh.modifiers[1]
			--create bones
			--format "dbump\n"
			--format "%" bonelist
			progressUpdate (100.0*11/16)
			--for i = 1 to bonelist.count do
			
			if loadrefpoints.tristate == 1 then
			(
				upto = header.totalbones
			)
			else
			(
				upto = header.wbone
			)
			format "Creating % bones\n" upto
			for i = 1 to upto do
			(
				--inverse bone transformation matrix
				transmatrix = ( inverse bonelist[i].RestPoseInverse )
				local bname = bonenames[i]
				if i > header.wbone then
				(
					if refpointsas.state == 1 then
					(
						--format "---\n%\n" pos
						newdummy = dummy name:bname
						newdummy.boxsize = [2,2,2]
						newdummy.pos = pos
					)
					else
					(
						length = 10 --should find childbone and make length distance between parent and child.
						bonenode = bonesys.createbone transmatrix.row4 (transmatrix.row4 + length*(normalize transmatrix.row3)) (normalize transmatrix.row2)
						--set bone name width etc
						bonenode.name = bname
						bonenode.width = 0.4
						bonenode.height = 0.4
					)
				)
				else
				(
					length = 10 --should find childbone and make length distance between parent and child.
					bonenode = bonesys.createbone transmatrix.row4 (transmatrix.row4 + length*(normalize transmatrix.row3)) (normalize transmatrix.row2)
					--Set bone name width etc
					bonenode.name = bname
					bonenode.width = 0.4
					bonenode.height = 0.4
					--Add to skinmod
					if i<upto do skinops.addbone skinmod node 0
					if i==upto do skinops.addbone skinmod node -1
				)
			)
			progressUpdate (100.0*12/16)
			--Fix progressupdate later.
			progressUpdate (100.0*13/16)
			--Weight verts
 			for i = 1 to new_mesh.verts.count do
			(
 				skinops.setvertexweights skinmod i boneweight[i] 1.0
 			)
			
			progressUpdate (100.0*14/16)
			format "Creating % links\n" upto
			for i = 1 to upto do
			(
				--get bone parent
				--print ""
				local parent = bonelist[i].parentboneindex
				local parentname = bonenames[i]
				--format "Parent: %, Name %" parent parentname
				if parent == 0 then
				(
					--do nothing, it's the parent
				)
				else
				(
					--link it to its parent
					node1 = getNodeByName bonenames[i]
					node2 = getNodeByName bonenames[parent]
					attachObjects node2 node1 move:false
					--attachObjects getNodeByName bonenames[i] getNodeByName bonenames[parent] move:false
				)
			)
			
 			progressUpdate (100.0*15/16)
			--fix model rotation
			fixrotnode = getNodeByName bonenames[1]
			if fixrotcheck.tristate == 1 do	
			(
				rotate fixrotnode 90 x_axis
				--switch y and z pos axis of root bone
				fixrotnodepos = fixrotnode.pos
				fixrotnode.pos = [fixrotnodepos[1],fixrotnodepos[3],fixrotnodepos[2]]
			)
			progressUpdate (100.0)
			progressEnd()
		),

		fn LoadModel=
		(
			progressUpdate (0.0)
			-- Load SCM_Header
			fseek bstream 0 #seek_set
			
			header = SCM_Header()
			header.LoadFromStream bstream
			--header.DebugOut()
			progressUpdate (100.0*2/16)
			if not header.IsValid() then
			(
				format "Invalid header, aborting.\n"
				return false;
			)
			progressUpdate (100.0*3/16)
			-- Load BONES
			fseek bstream 64 #seek_set
			bonestring = ReadFixedString bstream ( header.boneoffset - 64 ) false
			tokens = filterString bonestring "\t"

			--copy all tokens except the last one into bonenames array. Last token is fourcc SKEL.
			
			for i = 1 to ( tokens.count - 1 )do
			(
				--format "Bone %: %\n" i tokens[i] 
				append bonenames tokens[i]
			)
			--format "%" bonenames

			fseek bstream header.boneoffset #seek_set
			format "Loading bones \n"
			local curBone = SCM_Bone()
			for i = 1 to bonenames.count do
			(
				curBone.LoadFromStream bstream
				append bonelist (copy curBone)
				--format "\nBone %: %\n" i bonenames[i]
				--bonelist[i].DebugOut()
			)
			progressUpdate (100.0*4/16)

			fseek bstream header.vertoffset #seek_set
			format "Loading Vertices \n"
			local curVert = SCM_vert()
			for i = 1 to header.vertcount do
			(
				curVert.LoadFromStream bstream
				append vertlist (copy curVert)
				--get bone link from vert
				bone = ( curVert.boneindex + 1 )
				bone = bone[1]
				append boneweight bone
			)
			progressUpdate (100.0*5/16)
			
			fseek bstream header.indexoffset #seek_set
			format "Loading Tri index \n"
			local curTris = SCM_Tris()
			for i = 1 to header.tricount do
			(
				curTris.LoadFromStream bstream
				append trislist (copy curTris)
			)
			progressUpdate (100.0*6/16)
			
			CreateModel()
		)
	)
	--UI Stuff
	on importButton pressed do
	(
		local arghstrings = scriptsPath + "scmtool.ini"
		local filename = getINISetting arghstrings "settings" "file"
	    --setINISetting <filename_string> <section_string> <key_string> <key_value_string
		--setINISetting scriptsPath + "scmimp.ini" "settings" "file"
		local scmFilename = getOpenFileName caption:"Select SCM file" \
					filename:filename \
					types:"SupCom SCM (*.scm)|*.scm|All Files (*.*)|*.*|"

					--filename:(GetINISetting "" + scriptsPath + "scmpimp.ini" "settings" "file" ) \
		
		if scmFilename != undefined then
		(
			if DoesFileReallyExist scmFilename then
			(
				progressStart ("Loading Model")
				--try
				--(
					set coordsys world
					local loader = SCM_Loader()
					loader.Open scmFilename
					progressUpdate (100.0*1/16)
					loader.LoadModel()
					loader.Close()
					setINISetting arghstrings "settings" "file" scmFilename
					--format "OK\n"
				--)
				--catch
				--(
				--	progressEnd()
				--	throw
				--)
				
				max views redraw
			)
		)
	)

	
)


